from typing import Union, Optional, Any, Iterable, Mapping
from abc import ABCMeta, abstractmethod
import json


class JSONSerializable(metaclass=ABCMeta):

    """Common functionality for json serializable objects."""

    serialize = staticmethod(json.dumps)
    deserialize = staticmethod(json.loads)

    @abstractmethod
    def __str__(self):
        raise NotImplementedError()

    @classmethod
    def from_str(cls, s: str):
        payload = cls.deserialize(s)

        if not isinstance(payload, dict):
            raise ValueError("Payload should be dict")

        return cls(**payload)


class JSONRPC20Request(JSONSerializable):
    def __init__(self,
                method: str,
                params: Optional[Union[Mapping[str, Any], Iterable[Any]]] = None,
                id: Optional[Union[str, int]] = None,
                is_notification: bool = False
                ) -> None:

        self.payload = {
            'jsonrpc': '2.0'
        }
        self.method = method
        self.params = params
        self.id = id

    def __repr__(self):
        return repr(self.payload)

    def __str__(self):
        return self.serialize(self.payload)

    @property
    def method(self) -> str:
        return self.payload['method']

    @method.setter
    def method(self, value: str) -> None:
        if not isinstance(value, str):
            raise ValueError('"method" has to be str')
        self.payload['method'] = value

    @property
    def params(self) -> Optional[Union[Mapping[str, Any], Iterable[Any]]]:
        return self.payload.get('params')

    @params.setter
    def params(self, value: Optional[Union[Mapping[str, Any], Iterable[Any]]]) -> None:
        self.payload['params'] = value

    @property
    def id(self):
        if not self.is_notification:
            return self.payload.get('id')

    @id.setter
    def id(self, value: Optional[Union[str, int]]) -> None:
        self.payload['id'] = value

    @id.deleter
    def id(self):
        del self.payload['id']

    @property
    def is_notification(self):
        return 'id' not in self.payload

    @property
    def args(self):
        """ Method position arguments.
        :return tuple args: method position arguments.
        """
        # if not none and not mapping
        return tuple(self.params) if isinstance(self.params, list) else ()

    @property
    def kwargs(self):
        """ Method named arguments.
        :return dict kwargs: method named arguments.
        """
        # if mapping
        return self.params if isinstance(self.params, dict) else {}

class JSONRPCError(JSONSerializable):

    """ Error for JSON-RPC communication.
    When a rpc call encounters an error, the Response Object MUST contain the
    error member with a value that is a Object with the following members:
    Parameters
    ----------
    code: int
        A Number that indicates the error type that occurred.
        This MUST be an integer.
        The error codes from and including -32768 to -32000 are reserved for
        pre-defined errors. Any code within this range, but not defined
        explicitly below is reserved for future use. The error codes are nearly
        the same as those suggested for XML-RPC at the following
        url: http://xmlrpc-epi.sourceforge.net/specs/rfc.fault_codes.php
    message: str
        A String providing a short description of the error.
        The message SHOULD be limited to a concise single sentence.
    data: int or str or dict or list, optional
        A Primitive or Structured value that contains additional
        information about the error.
        This may be omitted.
        The value of this member is defined by the Server (e.g. detailed error
        information, nested errors etc.).
    """

    def __init__(self, code: int = None, message: str = None, data: Any = None):
        self.payload = {}
        self.code = getattr(self.__class__, "CODE", code)
        self.message = getattr(self.__class__, "MESSAGE", message)

        if data is not None:
            # NOTE: if not set in constructor, do not add 'data' to payload.
            # If data = null is requred, set it after object initialization.
            self.data = data

    def __repr__(self):
        return repr(self.payload)

    def __str__(self):
        return self.serialize(self.payload)

    @property
    def code(self):
        return self.payload["code"]

    @code.setter
    def code(self, value: int):
        if not isinstance(value, int):
            raise ValueError("Error code should be integer")

        self.payload["code"] = value

    @property
    def message(self):
        return self.payload["message"]

    @message.setter
    def message(self, value: str):
        if not isinstance(value, str):
            raise ValueError("Error message should be string")

        self.payload["message"] = value

    @property
    def data(self):
        return self.payload.get("data")

    @data.setter
    def data(self, value):
        self.payload["data"] = value

    @data.deleter
    def data(self):
        del self.payload["data"]


class JSONRPCParseError(JSONRPCError):

    """ Parse Error.
    Invalid JSON was received by the server.
    An error occurred on the server while parsing the JSON text.
    """

    CODE = -32700
    MESSAGE = "Parse error"


class JSONRPCInvalidRequest(JSONRPCError):

    """ Invalid Request.
    The JSON sent is not a valid Request object.
    """

    CODE = -32600
    MESSAGE = "Invalid Request"


class JSONRPCMethodNotFound(JSONRPCError):

    """ Method not found.
    The method does not exist / is not available.
    """

    CODE = -32601
    MESSAGE = "Method not found"


class JSONRPCInvalidParams(JSONRPCError):

    """ Invalid params.
    Invalid method parameter(s).
    """

    CODE = -32602
    MESSAGE = "Invalid params"


class JSONRPCInternalError(JSONRPCError):

    """ Internal error.
    Internal JSON-RPC error.
    """

    CODE = -32603
    MESSAGE = "Internal error"


class JSONRPCServerError(JSONRPCError):

    """ Server error.
    Reserved for implementation-defined server-errors.
    """

    CODE = -32000
    MESSAGE = "Server error"


class JSONRPC20Response(JSONSerializable):
    def __init__(self,
                result: Optional[Any] = None,
                error: Optional[JSONRPCError] = None,
                id: Optional[Union[int, str]] = None,
                ) -> None:

        self.payload = {
            'jsonrpc': '2.0'
        }

        if result is not None:
            self.payload['result'] = result

        if error is not None:
            self.payload['error'] = error

        self.id = id

    def __repr__(self):
        return repr(self.payload)

    def __str__(self):
        return self.serialize(self.payload)

    @property
    def id(self):
        return self.payload['id']

    @id.setter
    def id(self, value: Optional[Union[str, int]]) -> None:
        self.payload['id'] = value

    @property
    def result(self):
        return self.payload.get('result')

    @property
    def error(self):
        return self.payload.get('error')



import pytest

from ..core import JSONRPC20Request


class TestJSONRPC20Request:

    request_params = {
        "method": "add",
        "params": [1, 2],
        "id": 1,
    }

    def test_correct_init(self):
        JSONRPC20Request(**self.request_params)

    def test_validation_incorrect_no_parameters(self):
        with pytest.raises(TypeError):
            JSONRPC20Request()

    def test_default_not_notification(self):
        r = JSONRPC20Request(method='echo')
        assert not r.is_notification



import argparse
import asyncio
import json
import importlib.util
from inspect import getmembers, isfunction
from ajsonrpc import __version__
from ajsonrpc.dispatcher import Dispatcher
from ajsonrpc.manager import AsyncJSONRPCResponseManager
from ajsonrpc.core import JSONRPC20Request


class JSONRPCProtocol(asyncio.Protocol):
    def __init__(self, json_rpc_manager):
        self.json_rpc_manager = json_rpc_manager

    def connection_made(self, transport):
        self.transport = transport

    def handle_task_result(self, task):
        res = task.result()
        print(res)

        response_json = str(res).encode('utf-8')
        self.transport.write(response_json)

        print('Close the client socket')
        self.transport.close()

    def data_received(self, data):
        message = data.decode()
        print('Data received: {!r}\n'.format(message))

        request_method, request_message = message.split('\r\n', 1)
        if not request_method.startswith('POST'):
            print('Incorrect HTTP method, should be POST')

        headers, body = request_message.split('\r\n\r\n', 1)
        print('Body: {!r}\n'.format(body))

        # TODO: delegate to ResponseManager such string handling.
        data = json.loads(body)
        del data["jsonrpc"]
        req = JSONRPC20Request(**data)

        task = asyncio.async(self.json_rpc_manager.handle_request(req))
        task.add_done_callback(self.handle_task_result)


def main():
    """Usage: % -m examples.methods

    pipenv run async_json_rpc_server.py examples.methods

    """
    parser = argparse.ArgumentParser(
        add_help=True,
        description="Start async JSON-RPC 2.0 server")
    parser.add_argument(
        '--version', action='version',
        version='%(prog)s {version}'.format(version=__version__))
    parser.add_argument('module')

    args = parser.parse_args()

    spec = importlib.util.spec_from_file_location("module", args.module)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    # get functions from the module
    methods = getmembers(module, isfunction)
    print('Extracted methods: {}'.format(methods))
    dispatcher = Dispatcher(dict(methods))
    print('Dispatcher: {}'.format(dispatcher))

    json_rpc_manager = AsyncJSONRPCResponseManager(dispatcher=dispatcher)
    loop = asyncio.get_event_loop()
    # Each client connection will create a new protocol instance
    coro = loop.create_server(
        lambda: JSONRPCProtocol(json_rpc_manager),
        '127.0.0.1', 8888
    )
    server = loop.run_until_complete(coro)

    # Serve requests until Ctrl+C is pressed
    print('Serving on {}'.format(server.sockets[0].getsockname()))
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    # Close the server
    server.close()
    loop.run_until_complete(server.wait_closed())
    loop.close()


if __name__ == '__main__':
    main()



import inspect
from typing import Optional

from .core import JSONRPC20Response


class AsyncJSONRPCResponseManager:
    def __init__(self, dispatcher):
        self.dispatcher = dispatcher

    def handle(self, req):
        pass

    async def handle_request(self, request) -> Optional[JSONRPC20Response]:
        """Handle individual JSON-RPC request.
        
        Args:
            request (ajsonrpc.JSONRPC20Request) individual request
        
        Returns:
            ajsonrpc.JSONRPC20Response, optional json-rpc response or None if request is a notification
        
        """
        def make_response(**kwargs):
            response = JSONRPC20Response(id=request.id, **kwargs)
            response.request = request
            return response
        
        output = None
        try:
            method = self.dispatcher[request.method]
        except KeyError:
            # output = make_response(error=JSONRPCMethodNotFound()._data)
            pass
        else:
            try:
                result = method(*request.args, **request.kwargs) \
                    if not inspect.iscoroutinefunction(method) \
                    else await method(*request.args, **request.kwargs)
            # except JSONRPCDispatchException as e:
            #     output = make_response(error=e.error._data)
            except Exception as e:
                data = {
                    "type": e.__class__.__name__,
                    "args": e.args,
                    "message": str(e),
                }
                print(data)
                # if isinstance(e, TypeError) and is_invalid_params(method, *request.args, **request.kwargs):
                #     output = make_response(error=JSONRPCInvalidParams(data=data)._data)
                # else:
                #     output = make_response(error=JSONRPCServerError(data=data)._data)
            else:
                output = make_response(result=result)
        finally:
            # if not request.is_notification:
            return output


"""Method name to method mapper.

Dispatcher is a dict-like object which maps method_name to method.
For usage examples see :meth:`~Dispatcher.add_method`

"""
import functools
import collections


class Dispatcher(collections.MutableMapping):

    """Dictionary-like object which maps method_name to method."""

    def __init__(self, prototype=None):
        """ Build method dispatcher.

        Parameters
        ----------
        prototype : object or dict, optional
            Initial method mapping.

        Examples
        --------

        Init object with method dictionary.

        >>> Dispatcher({"sum": lambda a, b: a + b})
        None

        """
        self.method_map = dict()

        if prototype is not None:
            self.add_prototype(prototype)

    def __getitem__(self, key):
        return self.method_map[key]

    def __setitem__(self, key, value):
        self.method_map[key] = value

    def __delitem__(self, key):
        del self.method_map[key]

    def __len__(self):
        return len(self.method_map)

    def __iter__(self):
        return iter(self.method_map)

    def __repr__(self):
        return repr(self.method_map)

    @classmethod
    def extract_methods(cls, prototype, prefix=''):
        return dict(
            (prefix + method, getattr(prototype, method))
            for method in dir(prototype)
            if not method.startswith('_')
        )

    def add_prototype(self, prototype, prefix=''):
        mapping = self.extract_methods(prototype, prefix=prefix) \
            if not isinstance(prototype, collections.Mapping) \
            else prototype
        self.update(mapping)

    def add_class(self, cls, prefix=None):
        """Add class to dispatcher.

        Adds all of the public methods to dispatcher.

        Notes
        -----
            If class has instance methods (e.g. no @classmethod decorator),
            they likely would not work. Use :meth:`~add_object` instead.

        Parameters
        ----------
        cls : type
            class with methods to be added to dispatcher
        prefix : str, optional
            Method prefix. If not present, lowercased class name is used.

        """
        prefix = prefix or cls.__name__.lower() + '.'
        self.update(Dispatcher.extract_methods(cls), prefix=prefix)

    def add_object(self, obj, prefix=None):
        prefix = prefix or obj.__class__.__name__.lower() + '.'
        self.update(Dispatcher.extract_methods(obj), prefix=prefix)

    def add_method(self, f=None, name=None):
        """ Add a method to the dispatcher.

        Parameters
        ----------
        f : callable
            Callable to be added.
        name : str, optional
            Name to register (the default is function **f** name)

        Notes
        -----
        When used as a decorator keeps callable object unmodified.

        Examples
        --------

        Use as method

        >>> d = Dispatcher()
        >>> d.add_method(lambda a, b: a + b, name="sum")
        <function __main__.<lambda>>

        Or use as decorator

        >>> d = Dispatcher()
        >>> @d.add_method
            def mymethod(*args, **kwargs):
                print(args, kwargs)

        Or use as a decorator with a different function name
        >>> d = Dispatcher()
        >>> @d.add_method(name="my.method")
            def mymethod(*args, **kwargs):
                print(args, kwargs)

        """
        if name and not f:
            return functools.partial(self.add_method, name=name)

        self[name or f.__name__] = f
        return f



import pytest

from ..core import JSONRPC20Request
from ..manager import AsyncJSONRPCResponseManager


async def async_fun():
    return 'async function'


class TestJSONRPCResponseManager:
    dispatcher = {
        "add": sum,
        "multiply": lambda a, b: a * b,
        "list_len": len,
        "101_base": lambda **kwargs: int("101", **kwargs),
        "key_error": lambda: raise_(KeyError("error_explanation")),
        "type_error": lambda: raise_(TypeError("TypeError inside method")),
        "async_fun": async_fun,
    }
    manager = AsyncJSONRPCResponseManager(dispatcher=dispatcher)

    @pytest.mark.asyncio
    async def test_handle_request(self):
        req = JSONRPC20Request('add', params=[[1, 2]], id=0)
        res = await self.manager.handle_request(req)
        assert res.result == 3